# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import networkx as nx  # type: ignore

from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.suggestions.importers.do_all import (
    import_suggestion as import_doall,
)
from discopop_library.discopop_optimizer.suggestions.importers.reduction import (
    import_suggestion as import_reduction,
)


def import_suggestions(
    detection_result, graph: nx.DiGraph, get_next_free_node_id_function, environment: Experiment
) -> nx.DiGraph:
    """Imports the suggestions specified in res into the passed graph and returns the modified graph"""

    # import do-all
    for suggestion in detection_result.do_all:
        graph = import_doall(graph, suggestion, get_next_free_node_id_function, environment)

    # import reduction
    for suggestion in detection_result.reduction:
        graph = import_reduction(graph, suggestion, get_next_free_node_id_function, environment)
    return graph



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import os
import random
import string
import subprocess
import warnings
from typing import List, Tuple, Dict, cast, Optional

import jsonpickle  # type: ignore
import jsons  # type: ignore
import networkx as nx  # type: ignore

from discopop_explorer.PETGraphX import NodeID, PETGraphX
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.pattern_detectors.device_updates import DeviceUpdateInfo
from discopop_library.CodeGenerator.CodeGenerator import (
    from_pattern_info as code_gen_from_pattern_info,
)
from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.bindings.CodeStorageObject import CodeStorageObject
from discopop_library.discopop_optimizer.bindings.utilities import is_child_of_any
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.classes.system.devices.Device import Device
from discopop_library.discopop_optimizer.classes.system.devices.GPU import GPU
from discopop_library.discopop_optimizer.utilities.MOGUtilities import (
    data_at,
    get_parents,
)


def export_code(
    pet: PETGraphX,
    graph: nx.DiGraph,
    experiment: Experiment,
    cost_model: CostModel,
    context: ContextObject,
    label: str,
    parent_function: FunctionRoot,
):
    """Provides a binding to the discopop code generator and exports the code corresponding to the given cost model"""
    # only consider "empty", i.e. sequential cases, if they are either the sequential or the locally optimized option
    if len(cost_model.path_decisions) == 0:
        if label not in ["Sequential", "Locally Optimized"]:
            print("warning: skipped empty model in code export.")

    # collect suggestions to be applied
    suggestions: List[Tuple[Device, PatternInfo, str, Optional[int]]] = []
    for decision in cost_model.path_decisions:
        graph_node = data_at(graph, decision)
        device: Device = experiment.get_system().get_device(0 if graph_node.device_id is None else graph_node.device_id)
        if graph_node.suggestion is None:
            continue
        suggestion, suggestion_type = device.get_device_specific_pattern_info(
            graph_node.suggestion, graph_node.suggestion_type
        )
        suggestions.append((device, suggestion, suggestion_type, decision))

    # todo collect updates to be applied
    for update in context.necessary_updates:
        if update.source_device_id != update.target_device_id:
            # calculate correct position for update (updates inside target regions are NOT allowed by OpenMP!)
            # insert pragma before the usage position of the data
            unchecked_target_nodes: List[int] = [update.target_node_id]
            checked_target_nodes: List[int] = []

            while len(unchecked_target_nodes) != 0:
                cur_node = unchecked_target_nodes.pop()
                if is_child_of_any(
                    graph,
                    cur_node,
                    [s_node_id for _, sugg, s_type, s_node_id in suggestions if s_type.startswith("gpu_")],
                ):
                    tmp_parents = get_parents(graph, cur_node)
                    if len(tmp_parents) != 0:
                        unchecked_target_nodes += [
                            p for p in tmp_parents if p not in checked_target_nodes and p not in unchecked_target_nodes
                        ]

                else:
                    if cur_node not in checked_target_nodes:
                        checked_target_nodes.append(cur_node)

            for checked_node_id in checked_target_nodes:
                start_line = pet.node_at(cast(NodeID, data_at(graph, checked_node_id).original_cu_id)).start_position()
                end_line = start_line

                # register update
                source_cu_id = cast(NodeID, data_at(graph, update.source_node_id).original_cu_id)
                target_cu_id = cast(NodeID, data_at(graph, update.target_node_id).original_cu_id)

                if source_cu_id is None or target_cu_id is None:
                    warnings.warn("Could not register update: " + str(update) + " @ Line: " + start_line)
                else:
                    # get updated variable
                    var_obj = pet.get_variable(target_cu_id, cast(str, update.write_data_access.var_name))
                    if var_obj is None:
                        raise ValueError(
                            "Could not find variable object for: "
                            + str(update)
                            + "  ->  "
                            + str(update.write_data_access.var_name)
                        )

                    # get amount of elements targeted by the update
                    update_elements = int(
                        int(experiment.get_memory_region_size(update.write_data_access.memory_region)[0].evalf())
                        / var_obj.sizeInByte
                    )

                    # debug!
                    # add memory access to var_name
                    dbg_info = (
                        "-"
                        + str(update.write_data_access.memory_region)
                        + "-"
                        + str(update.write_data_access.unique_id)
                        + "-vartype:"
                        + var_obj.type
                    )

                    # add range to updated var name if necessary
                    if update_elements > 1 and update.write_data_access.var_name is not None and "**" in var_obj.type:
                        updated_var_name: Optional[str] = (
                            str(update.write_data_access.var_name)
                            + "[:"
                            + str(update_elements)
                            + "]"
                            # + dbg_info
                        )
                    else:
                        updated_var_name = update.write_data_access.var_name

                    suggestions.append(
                        (
                            experiment.get_system().get_device(update.source_device_id),
                            DeviceUpdateInfo(
                                pet,
                                pet.node_at(source_cu_id),
                                pet.node_at(target_cu_id),
                                update.write_data_access.memory_region,
                                updated_var_name,
                                0 if update.source_device_id is None else update.source_device_id,
                                0 if update.source_device_id is None else cast(int, update.target_device_id),
                                start_line,
                                end_line,
                                update.is_first_data_occurrence,
                                cast(GPU, experiment.get_system().get_device(update.source_device_id)).openmp_device_id,
                                cast(GPU, experiment.get_system().get_device(update.target_device_id)).openmp_device_id,
                            ),
                            "device_update",
                            None,
                        )
                    )

    # remove duplicates
    to_be_removed = []
    buffer = []
    for entry in suggestions:
        entry_str = str(entry[0]) + ";" + str(entry[1]) + ";" + str(entry[2]) + ";" + str(entry[3])
        if entry_str not in buffer:
            buffer.append(entry_str)
        else:
            to_be_removed.append(entry)
    for entry in to_be_removed:
        if entry in suggestions:
            suggestions.remove(entry)

    # remove updates, if an identical entry exists
    to_be_removed = []
    for entry_1 in suggestions:
        if not isinstance(entry_1[1], DeviceUpdateInfo):
            continue
        entry_str_1 = (
            str(entry_1[0])
            + ";"
            + (
                entry_1[1].get_str_without_first_data_occurrence()
                if isinstance(entry_1[1], DeviceUpdateInfo)
                else str(entry_1[1])
            )
            + ";"
            + str(entry_1[2])
            + ";"
            + str(entry_1[3])
        )
        for entry_2 in suggestions:
            if not isinstance(entry_1[1], DeviceUpdateInfo):
                continue
            entry_str_2 = (
                str(entry_2[0])
                + ";"
                + (
                    entry_2[1].get_str_without_first_data_occurrence()
                    if isinstance(entry_2[1], DeviceUpdateInfo)
                    else str(entry_2[1])
                )
                + ";"
                + str(entry_2[2])
                + ";"
                + str(entry_2[3])
            )
            if entry_str_1 == entry_str_2:
                if (
                    entry_1[1].is_first_data_occurrence
                    and not cast(DeviceUpdateInfo, entry_2[1]).is_first_data_occurrence
                ):
                    to_be_removed.append(entry_2)
    for entry in to_be_removed:
        if entry in suggestions:
            suggestions.remove(entry)

    if False:
        # todo cleanup
        for sugg in suggestions:
            if type(sugg[1]) == DeviceUpdateInfo:
                tmp = cast(DeviceUpdateInfo, sugg[1])
                print(
                    "First" if tmp.is_first_data_occurrence else "",
                    "Update: ",
                    tmp.start_line,
                    "@",
                    tmp.source_device_id,
                    "->",
                    tmp.start_line,
                    "@",
                    tmp.target_device_id,
                    "|",
                    tmp.mem_reg,
                    "|",
                    tmp.var_name,
                )

    # prepare patterns by type
    patterns_by_type: Dict[str, List[PatternInfo]] = dict()
    for device, pattern, s_type, s_node_id in suggestions:
        if s_type not in patterns_by_type:
            patterns_by_type[s_type] = []
        # add device id to pattern
        pattern.dp_optimizer_device_id = device.openmp_device_id
        patterns_by_type[s_type].append(pattern)

    # invoke the discopop code generator
    modified_code = code_gen_from_pattern_info(
        experiment.file_mapping,
        patterns_by_type,
        skip_compilation_check=True,  # False,
        compile_check_command=experiment.compile_check_command,
    )
    # create patches from the modified code
    patches = __convert_modified_code_to_patch(experiment, modified_code)

    # save modified code as CostStorageObject
    export_dir = os.path.join(experiment.discopop_optimizer_path, "code_exports")
    if not os.path.exists(export_dir):
        os.makedirs(export_dir)

    # generate unique hash name
    hash_name = "".join(random.choices(string.ascii_uppercase + string.digits, k=8))
    while os.path.exists(os.path.join(export_dir, hash_name)):
        hash_name = "".join(random.choices(string.ascii_uppercase + string.digits, k=8))

    code_storage = CodeStorageObject(cost_model, patches, parent_function, hash_name, label)

    # export code_storage object to json
    print("Export JSON TO: ", os.path.join(export_dir, hash_name + ".json"))
    with open(os.path.join(export_dir, hash_name + ".json"), "w+") as f:
        f.write(jsonpickle.dumps(code_storage))
        f.flush()
        f.close()

    print("\n############################")
    print("Modified Code: Decisions: ", cost_model.path_decisions)
    print("Exporting to: ", export_dir)
    print("############################")
    for file_id in patches:
        print("#### File ID: ", file_id, " ####\n")
        print(patches[file_id])
        print("\n")


def __convert_modified_code_to_patch(experiment: Experiment, modified_code: Dict[int, str]) -> Dict[int, str]:
    patches: Dict[int, str] = dict()
    hash_name = "".join(random.choices(string.ascii_uppercase + string.digits, k=8))
    tmp_file_name = os.path.join(os.getcwd(), hash_name + ".tmp")
    for file_id in modified_code:
        # write modified code to file
        with open(tmp_file_name, "w+") as f:
            f.write(modified_code[file_id])
            f.flush()
            f.close()

        # generate diff
        diff_name = tmp_file_name + ".diff"
        command = [
            "diff",
            "-Naru",
            str(experiment.file_mapping[file_id]),
            tmp_file_name,
        ]
        result = subprocess.run(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            cwd=os.getcwd(),
        )
        if result.returncode != 0:
            print("RESULT: ", result.returncode)
            print("STDERR:")
            print(result.stderr)
            print("STDOUT: ")
            print(result.stdout)

        # save diff
        patches[file_id] = result.stdout

        # cleanup environment
        if os.path.exists(tmp_file_name):
            os.remove(tmp_file_name)
        if os.path.exists(diff_name):
            os.remove(diff_name)

    return patches



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

"""Discopop Suggestion Optimizer

Usage:
    discopop_optimizer --compile-command <str> [--project <path>] [--file-mapping <path>] [--detection-result-dump <path>]
        [--execute-created-models] [--execute-single-model <path>] [--clean-created-code] [--code-export-path <path>] [--dp-output-path <path>]
        [--executable-arguments <string>] [--linker-flags <string>] [--make-target <string>] [--make-flags <string>]
        [--executable-name <string>] [--execution-repetitions <int>] [--execution-append-measurements]
        [--exhaustive-search] [--headless-mode] [--doall-microbench-file <string>] [--reduction-microbench-file <string>]

OPTIONAL ARGUMENTS:
    --project=<path>            Path to the directory that contains your makefile [default: .]
    --file-mapping=<path>       Path to the FileMapping.txt. [default: FileMapping.txt]
    --detection-result-dump=<path>  Path to the dumped detection result JSON. [default: explorer/detection_result_dump.json]
    --execute-created-models    Compiles, executes and measures models already stored in the project folder.
                                Does not start the optimization pipeline.
                                Required: --executable-name
    --execute-single-model=<path>       Execute the given model only. Path to the JSON or filename, if the model is located in --code-export-path.
                                        Does not start the optimization pipeline.
                                        Required: --executable-name [default: ]
    --execution-repetitions=<int>       Repeat measurements [default: 1]
    --execution-append-measurements     If set, measurement files from previous executions will be kept and new results
                                        will be appended.
    --clean-created-code        Removes all stored code modifications.
    --code-export-path=<path>   Directory where generated CodeStorageObjects are located. [default: optimizer/code_exports]
    --dp-output-path=<path>     Directory where output files of DiscoPoP are located. [default: .]
    --executable-name=<string>  Name of the executable generate by your makefile.
                                Must be specified if --execute-created-models is used! [default: ]
    --executable-arguments=<string>
                                run your application with these arguments [default: ]
    --linker-flags=<string>     if your build requires to link other libraries
                                please provide the necessary linker flags. e.g. -lm [default: ]
    --make-target=<string>      specify a specific Make target to be built,
                                if not specified the default make target is used. [default: ]
    --make-flags=<string>       specify flags which will be passed through to make. [default: ]
    --compile-command=<string>    specify a command which shall be executed to compile the application
    --exhaustive-search         Perform an exhaustive search for the optimal set of parallelization suggestions for the
                                given environment
    --headless-mode             Do not show any GUI prompts. Does not reuse prior results.
                                Uses the suggested default values. Creates random models.
                                Exports all models to code. Saves all models.
    --reduction-microbench-file=<string>     Path to the microbenchmark output which represents the overhead
                                             of reduction suggestions.
                                             Default model: 0
    --doall-microbench-file=<string>         Path to the microbenchmark output which represents the overhead
                                             of reduction suggestions.
                                             Default model: 0
    -h --help                   Show this screen
"""
import os
import shutil
import sys
import tkinter as tk
from typing import Optional, Union

import jsonpickle  # type: ignore
import pstats2  # type:ignore
from docopt import docopt  # type:ignore
from schema import Schema, Use, SchemaError  # type:ignore

from discopop_library.result_classes.DetectionResult import DetectionResult
from discopop_library.discopop_optimizer.Microbench.ExtrapInterpolatedMicrobench import (
    ExtrapInterpolatedMicrobench,
)
from discopop_library.discopop_optimizer.Microbench.Microbench import MicrobenchType
from discopop_library.discopop_optimizer.OptimizationGraph import OptimizationGraph
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.Variables.ExperimentUtils import (
    restore_session,
    show_function_models,
    export_to_json,
)
from discopop_library.discopop_optimizer.classes.system.System import System
from discopop_library.discopop_optimizer.classes.system.devices.CPU import CPU
from discopop_library.discopop_optimizer.classes.system.devices.GPU import GPU
from discopop_library.discopop_optimizer.execution.stored_models import (
    execute_stored_models,
    execute_single_model,
)
import tkinter.messagebox

docopt_schema = Schema(
    {
        "--project": Use(str),
        "--file-mapping": Use(str),
        "--detection-result-dump": Use(str),
        "--execute-created-models": Use(str),
        "--clean-created-code": Use(str),
        "--code-export-path": Use(str),
        "--dp-output-path": Use(str),
        "--executable-arguments": Use(str),
        "--executable-name": Use(str),
        "--linker-flags": Use(str),
        "--make-target": Use(str),
        "--make-flags": Use(str),
        "--execution-repetitions": Use(str),
        "--execute-single-model": Use(str),
        "--compile-command": Use(str),
        "--execution-append-measurements": Use(str),
        "--exhaustive-search": Use(str),
        "--headless-mode": Use(str),
        "--doall-microbench-file": Use(str),
        "--reduction-microbench-file": Use(str),
    }
)


def get_path(base_path: str, file_name: str) -> str:
    """Combines path and filename if it is not absolute

    :param base_path: path
    :param file_name: file name
    :return: path to file
    """
    result_path = file_name if os.path.isabs(file_name) else os.path.join(base_path, file_name)
    return os.path.normpath(result_path)


def main():
    """Invokes the discopop_optimizer using the given parameters"""
    arguments = docopt(__doc__)

    try:
        arguments = docopt_schema.validate(arguments)
    except SchemaError as e:
        exit(e)

    # prepare arguments
    arguments["--project"] = get_path(os.getcwd(), arguments["--project"])
    arguments["--execute-created-models"] = False if arguments["--execute-created-models"] == "False" else True
    arguments["--execution-append-measurements"] = (
        False if arguments["--execution-append-measurements"] == "False" else True
    )
    arguments["--clean-created-code"] = False if arguments["--clean-created-code"] == "False" else True
    arguments["--code-export-path"] = get_path(arguments["--project"], arguments["--code-export-path"])
    arguments["--exhaustive-search"] = False if arguments["--exhaustive-search"] == "False" else True
    arguments["--headless-mode"] = False if arguments["--headless-mode"] == "False" else True
    arguments["--dp-output-path"] = get_path(arguments["--project"], arguments["--dp-output-path"])
    arguments["--file-mapping"] = get_path(arguments["--dp-output-path"], arguments["--file-mapping"])
    arguments["--detection-result-dump"] = get_path(arguments["--dp-output-path"], arguments["--detection-result-dump"])
    arguments["--dp-optimizer-path"] = os.path.join(arguments["--project"], ".discopop_optimizer")
    arguments["--make-target"] = None if arguments["--make-target"] == "None" else arguments["--make-target"]
    arguments["--execute-single-model"] = (
        None
        if len(arguments["--execute-single-model"]) == 0
        else get_path(arguments["--code-export-path"], arguments["--execute-single-model"])
    )
    arguments["--compile-command"] = (
        None if len(arguments["--compile-command"]) == 0 else arguments["--compile-command"]
    )

    print("Starting discopop_optimizer...")
    for arg_name in arguments:
        print("\t", arg_name, "=", arguments[arg_name])

    if arguments["--execute-created-models"] or arguments["--execute-single-model"] is not None:
        if arguments["--executable-name"] == "":
            raise ValueError("Please specify the name of your executable using --executable-name!")
        if arguments["--execute-single-model"] is None:
            execute_stored_models(arguments)
        else:
            execute_single_model(arguments)
        sys.exit(0)

    if arguments["--clean-created-code"]:
        # remove stored parallel code if requested
        print("Removing old exported code...", end="")
        if os.path.exists(arguments["--code-export-path"]):
            shutil.rmtree(arguments["--code-export-path"])
        os.makedirs(arguments["--code-export-path"])
        print("Done.")

    start_optimizer(arguments)


def start_optimizer(arguments, parent_frame: Optional[tk.Frame] = None):
    # create gui frame if none given
    if parent_frame is None:
        tk_root: Union[tk.Tk, tk.Toplevel] = tk.Tk()
        # configure window size
        tk_root.geometry("1000x600")
        parent_frame = tk.Frame(tk_root)
        parent_frame.pack(fill=tk.BOTH)
        destroy_window_after_execution = True
    else:
        tk_root = parent_frame.winfo_toplevel()
        destroy_window_after_execution = False

    # ask if previous session should be loaded
    load_result: bool = False
    if (
        os.path.exists(os.path.join(arguments["--dp-optimizer-path"], "last_experiment.pickle"))
        and not arguments["--headless-mode"]
    ):
        load_result = tkinter.messagebox.askyesno(
            parent=parent_frame,
            title="Restore Results?",
            message="Do you like to load the experiment from the previous session?",
        )
    if load_result:
        # load results from previous session
        experiment = restore_session(os.path.join(arguments["--dp-optimizer-path"], "last_experiment.pickle"))
        show_function_models(experiment, parent_frame, destroy_window_after_execution)
        # save experiment to disk
        export_to_json(experiment)

    else:
        # create a new session
        # load detection result
        print("Loading detection result and PET...", end="")
        detection_result_dump_str = ""
        with open(arguments["--detection-result-dump"], "r") as f:
            detection_result_dump_str = f.read()
        detection_result: DetectionResult = jsonpickle.decode(detection_result_dump_str)
        print("Done")

        # define System
        # todo make system user-configurable, or detect it using a set of benchmarks
        system = System()

        # todo connections between devices might happen as update to host + update to second device.
        #  As of right now, connections between two devices are implemented in this manner.
        # todo check if OpenMP allows direct data transfers between devices

        # load overhead measurements into system if existent
        if arguments["--doall-microbench-file"] != "None":
            microbench_file = arguments["--doall-microbench-file"]
            if not os.path.isfile(microbench_file):
                raise FileNotFoundError(f"Microbenchmark file not found: {microbench_file}")
            # construct and set overhead model for doall suggestions
            system.set_device_doall_overhead_model(
                system.get_device(0),
                ExtrapInterpolatedMicrobench(microbench_file).getFunctionSympy(),
            )
        if arguments["--reduction-microbench-file"] != "None":
            microbench_file = arguments["--reduction-microbench-file"]
            if not os.path.isfile(microbench_file):
                raise FileNotFoundError(f"Microbenchmark file not found: {microbench_file}")
            # construct and set overhead model for reduction suggestions
            system.set_reduction_overhead_model(
                system.get_device(0),
                ExtrapInterpolatedMicrobench(microbench_file).getFunctionSympy(benchType=MicrobenchType.FOR),
            )

        # define Environment
        experiment = Experiment(
            arguments["--project"],
            arguments["--dp-output-path"],
            arguments["--dp-optimizer-path"],
            arguments["--code-export-path"],
            arguments["--file-mapping"],
            system,
            detection_result,
            arguments,
        )

        # invoke optimization graph
        optimization_graph = OptimizationGraph(
            arguments["--dp-output-path"],
            experiment,
            arguments,
            parent_frame,
            destroy_window_after_execution,
        )

    if destroy_window_after_execution:
        try:
            tk_root.destroy()
        except tkinter.TclError:
            # Window has been destroyed already
            pass


if __name__ == "__main__":
    main()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Dict

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot


class CodeStorageObject(object):
    cost_model: CostModel
    patches: Dict[int, str]
    parent_function: FunctionRoot
    model_id: str
    label: str

    def __init__(
        self,
        cost_model: CostModel,
        patches: Dict[int, str],
        parent_function: FunctionRoot,
        model_id: str,
        label: str,
    ):
        self.cost_model = cost_model
        self.patches = patches
        self.parent_function = parent_function
        self.model_id = model_id
        self.label = label



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import copy
from typing import Dict, List, Tuple, Set

import networkx as nx  # type: ignore

from discopop_explorer.PETGraphX import (
    PETGraphX,
    FunctionNode,
    EdgeType,
    LoopNode,
    CUNode,
    NodeID,
    MemoryRegion,
)
from discopop_explorer.utils import calculate_workload
from discopop_library.discopop_optimizer.PETParser.DataAccesses.FromCUs import (
    get_data_accesses_for_cu,
)
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.nodes.ContextMerge import ContextMerge
from discopop_library.discopop_optimizer.classes.nodes.ContextRestore import ContextRestore
from discopop_library.discopop_optimizer.classes.nodes.ContextSave import ContextSave
from discopop_library.discopop_optimizer.classes.nodes.ContextSnapshot import ContextSnapshot
from discopop_library.discopop_optimizer.classes.nodes.ContextSnapshotPop import ContextSnapshotPop
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.classes.nodes.Loop import Loop
from discopop_library.discopop_optimizer.classes.nodes.Workload import Workload
from discopop_library.discopop_optimizer.utilities.MOGUtilities import (
    data_at,
    get_successors,
    get_children,
    add_successor_edge,
    add_child_edge,
    add_temporary_edge,
    redirect_edge,
    convert_temporary_edges,
    get_all_function_nodes,
    get_read_and_written_data_from_subgraph,
)


class PETParser(object):
    pet: PETGraphX
    graph: nx.DiGraph
    next_free_node_id: int
    cu_id_to_graph_node_id: Dict[NodeID, int]
    experiment: Experiment
    in_data_flow: Dict[int, Set[int]]
    out_data_flow: Dict[int, Set[int]]

    def __init__(self, pet: PETGraphX, experiment: Experiment):
        self.pet = pet
        self.graph = nx.DiGraph()
        self.next_free_node_id = 0
        self.cu_id_to_graph_node_id = dict()
        self.experiment = experiment

    def parse(self) -> Tuple[nx.DiGraph, int]:
        self.__add_cu_nodes()
        self.__add_functions()
        self.__add_pet_successor_edges()
        self.__add_loop_nodes()

        self.__parse_branched_sections()
        convert_temporary_edges(self.graph)

        self.__mark_branch_affiliation()
        self.__calculate_data_flow()

        return self.graph, self.next_free_node_id

    def get_new_node_id(self) -> int:
        """returns the next free node id and increments the counter"""
        buffer = self.next_free_node_id
        self.next_free_node_id += 1
        return buffer

    def __parse_branched_sections(self):
        """Branched sections in the CU Graph are represented by a serialized version in the MOG.
        To make this possible, Context Snapshot, Restore and Merge points are added to allow a synchronization
        'between' the different branches"""
        visited_nodes: Set[int] = set()
        for function_node in self.pet.all_nodes(FunctionNode):
            _, _ = self.__parse_raw_node(self.cu_id_to_graph_node_id[function_node.id], visited_nodes)

        # remove visited nodes, since duplicates exist now
        for node_id in visited_nodes:
            self.graph.remove_node(node_id)

    def __parse_raw_node(self, root_node_id: int, visited_nodes: Set[int]) -> Tuple[int, int]:
        """parses the nodes starting from the root node and returns the id of the first and last node in the created linearized path"""
        visited_nodes.add(root_node_id)
        # if this node has children, parse them before proceeding
        children_entry_points: List[int] = []
        for child_id in get_children(self.graph, root_node_id):
            child_entry_point, _ = self.__parse_raw_node(child_id, visited_nodes)
            children_entry_points.append(child_entry_point)

        if len(get_successors(self.graph, root_node_id)) <= 1:
            return self.__parse_path_node(root_node_id, children_entry_points, visited_nodes)
        else:
            return self.__parse_branching_point(root_node_id, children_entry_points, visited_nodes)

    def __parse_path_node(
        self, root_node_id: int, connect_to_children: List[int], visited_nodes: Set[int]
    ) -> Tuple[int, int]:
        """parses a regular path node, i.e. one with at most one successor node.
        returns the id of the first and last node in the created linearized path"""

        # create a duplicate of the root node
        duplicate_node_id = self.get_new_node_id()
        tmp_node_data = data_at(self.graph, root_node_id)
        self.graph.add_node(duplicate_node_id, data=tmp_node_data)

        # connect duplicated entry node to children
        for children_id in connect_to_children:
            add_child_edge(self.graph, duplicate_node_id, children_id)

        if len(get_successors(self.graph, root_node_id)) == 0:
            # root is a path end
            return duplicate_node_id, duplicate_node_id
        else:
            # proceed to successor
            successor_id = get_successors(self.graph, root_node_id)[0]
            path_entry, path_end = self.__parse_raw_node(successor_id, visited_nodes)

            # connect created duplicate to the successor
            add_temporary_edge(self.graph, duplicate_node_id, path_entry)

            return duplicate_node_id, path_end

    def __parse_branching_point(
        self, root_node_id: int, connect_to_children: List[int], visited_nodes: Set[int]
    ) -> Tuple[int, int]:
        """parses a branching point, i.e. a node with more than one successor.
        returns the id of the first and last node in the created linearized path"""

        #########
        # Step 1: create duplicate of root node
        # Step 2: create and connect context snapshot
        # Step 3: parse branches
        #   Step 3.1: create and connect context restore node
        #   Step 3.2: parse branch
        #   Step 3.3: create context save node
        #   Step 3.3: connect restore and save not to branch entry and exit
        # step 4: create and connect context merge node
        # Step 5: create and connect context snapshot pop
        ########

        # Step 1: create duplicate of root node
        duplicate_node_id = self.get_new_node_id()
        tmp_node_data = data_at(self.graph, root_node_id)
        self.graph.add_node(duplicate_node_id, data=tmp_node_data)

        # Step 2: create and connect context snapshot
        context_snapshot_id = self.get_new_node_id()
        self.graph.add_node(context_snapshot_id, data=ContextSnapshot(context_snapshot_id, self.experiment))
        add_temporary_edge(self.graph, duplicate_node_id, context_snapshot_id)

        # Step 3: parse branches
        last_added_node_id = context_snapshot_id
        for successor in get_successors(self.graph, root_node_id):
            # Step 3.1: create and connect context restore node
            branch_context_restore_id = self.get_new_node_id()
            self.graph.add_node(
                branch_context_restore_id,
                data=ContextRestore(branch_context_restore_id, self.experiment),
            )
            add_temporary_edge(self.graph, last_added_node_id, branch_context_restore_id)

            # Step 3.2: parse branch
            branch_entry, branch_exit = self.__parse_raw_node(successor, visited_nodes)

            # Step 3.3: create context save node
            branch_context_save_id = self.get_new_node_id()
            self.graph.add_node(branch_context_save_id, data=ContextSave(branch_context_save_id, self.experiment))

            # Step 3.3: connect restore and save node to branch entry and exit
            add_temporary_edge(self.graph, branch_context_restore_id, branch_entry)
            add_temporary_edge(self.graph, branch_exit, branch_context_save_id)

            # update last_added_node_id
            last_added_node_id = branch_context_save_id

        # step 4: create and connect context merge node
        context_merge_node_id = self.get_new_node_id()
        self.graph.add_node(context_merge_node_id, data=ContextMerge(context_merge_node_id, self.experiment))
        add_temporary_edge(self.graph, last_added_node_id, context_merge_node_id)

        # Step 5: create and connect context snapshot pop
        context_snapshot_pop_id = self.get_new_node_id()
        self.graph.add_node(
            context_snapshot_pop_id,
            data=ContextSnapshotPop(context_snapshot_pop_id, self.experiment),
        )
        add_temporary_edge(self.graph, context_merge_node_id, context_snapshot_pop_id)

        # connect duplicated entry node to children
        for children_id in connect_to_children:
            add_child_edge(self.graph, duplicate_node_id, children_id)

        return duplicate_node_id, context_snapshot_pop_id

    def __add_cu_nodes(self):
        """adds Workload nodes which represent the CU Nodes to the graph.
        The added nodes will not be connected in any way."""
        for cu_node in self.pet.all_nodes(CUNode):
            # create new node for CU
            new_node_id = self.get_new_node_id()
            self.cu_id_to_graph_node_id[cu_node.id] = new_node_id
            # calculate accessed data
            written_memory_regions, read_memory_regions = get_data_accesses_for_cu(self.pet, cu_node.id)
            self.graph.add_node(
                new_node_id,
                data=Workload(
                    node_id=new_node_id,
                    experiment=self.experiment,
                    cu_id=cu_node.id,
                    sequential_workload=0,
                    parallelizable_workload=calculate_workload(
                        self.pet, cu_node, ignore_function_calls_and_cached_values=True
                    ),
                    written_memory_regions=written_memory_regions,
                    read_memory_regions=read_memory_regions,
                ),
            )

    def __add_loop_nodes(self):
        """adds Loop Nodes to the graph.
        connects contained nodes using Children edges"""
        for loop_node in self.pet.all_nodes(LoopNode):
            # calculate metadata
            loop_subtree = self.pet.subtree_of_type(loop_node, CUNode)

            # get iteration count for loop
            if loop_node.loop_data is not None:
                iteration_count = max(1, loop_node.loop_data.average_iteration_count)
            else:
                # loop has not been executed, thus, no information was gathered during the profiling
                # set iteration_count to prevent divisions by zero
                iteration_count = 1

            # create new node for Loop
            new_node_id = self.get_new_node_id()
            self.cu_id_to_graph_node_id[loop_node.id] = new_node_id
            self.graph.add_node(
                new_node_id,
                data=Loop(
                    node_id=new_node_id,
                    cu_id=loop_node.id,
                    discopop_workload=calculate_workload(
                        self.pet, loop_node, ignore_function_calls_and_cached_values=True
                    ),
                    iterations=iteration_count,
                    position=loop_node.start_position(),
                    experiment=self.experiment,
                ),
            )
            # connect loop node and entry node via a child edge
            entry_node_cu_id = loop_node.get_entry_node(self.pet).id
            add_child_edge(self.graph, new_node_id, self.cu_id_to_graph_node_id[entry_node_cu_id])

            # redirect edges from outside the loop to the entry node to the Loop node
            for s, t, d in self.pet.in_edges(entry_node_cu_id, EdgeType.SUCCESSOR):
                if self.pet.node_at(s) not in loop_subtree:
                    redirect_edge(
                        self.graph,
                        old_source_id=self.cu_id_to_graph_node_id[s],
                        new_source_id=self.cu_id_to_graph_node_id[s],
                        old_target_id=self.cu_id_to_graph_node_id[entry_node_cu_id],
                        new_target_id=new_node_id,
                    )

            # redirect edges to the outside of the loop
            for s, t, d in self.pet.out_edges(entry_node_cu_id, EdgeType.SUCCESSOR):
                if self.pet.node_at(t) not in loop_subtree:
                    redirect_edge(
                        self.graph,
                        old_source_id=self.cu_id_to_graph_node_id[entry_node_cu_id],
                        new_source_id=new_node_id,
                        old_target_id=self.cu_id_to_graph_node_id[t],
                        new_target_id=self.cu_id_to_graph_node_id[t],
                    )

            # copy entry node
            entry_node_id = self.cu_id_to_graph_node_id[entry_node_cu_id]
            entry_node_data = data_at(self.graph, entry_node_id)
            copied_entry_node_id = self.get_new_node_id()
            self.graph.add_node(copied_entry_node_id, data=entry_node_data)

            # redirect edges from inside the loop to the copy of the entry node
            for s, t, d in self.pet.in_edges(entry_node_cu_id, EdgeType.SUCCESSOR):
                if self.pet.node_at(s) in loop_subtree:
                    redirect_edge(
                        self.graph,
                        old_source_id=self.cu_id_to_graph_node_id[s],
                        new_source_id=self.cu_id_to_graph_node_id[s],
                        old_target_id=self.cu_id_to_graph_node_id[entry_node_cu_id],
                        new_target_id=copied_entry_node_id,
                    )

            # redirect accesses to the cu_id of the entry node to the newly created loop node
            self.cu_id_to_graph_node_id[entry_node_cu_id] = new_node_id

    def __add_functions(self):
        """parse function nodes in the PET graph.
        Results in the creation of a forest of function graphs."""
        for function_node in self.pet.all_nodes(FunctionNode):
            # create function root node and register it in the graph
            new_node_id = self.get_new_node_id()
            self.graph.add_node(
                new_node_id,
                data=FunctionRoot(
                    node_id=new_node_id,
                    experiment=self.experiment,
                    cu_id=function_node.id,
                    name=function_node.name,
                ),
            )
            # connect function node to its entry node
            add_child_edge(
                self.graph,
                new_node_id,
                self.cu_id_to_graph_node_id[function_node.get_entry_cu_id(self.pet)],
            )
            # save ID
            self.cu_id_to_graph_node_id[function_node.id] = new_node_id

    def __add_pet_successor_edges(self):
        for cu_node in self.pet.all_nodes(CUNode):
            for successor_cu_id in [t for s, t, d in self.pet.out_edges(cu_node.id, EdgeType.SUCCESSOR)]:
                add_successor_edge(
                    self.graph,
                    self.cu_id_to_graph_node_id[cu_node.id],
                    self.cu_id_to_graph_node_id[successor_cu_id],
                )

    def __mark_branch_affiliation(self):
        """Mark each nodes' branch affiliation to allow a simple check for 'on same branch' relation
        without considering the successor relation."""

        def mark_branched_section(node, branch_stack):
            node_data = data_at(self.graph, node)
            if isinstance(node_data, ContextSnapshot):
                branch_stack.append(node)
            elif isinstance(node_data, ContextSnapshotPop):
                branch_stack.pop()
            # mark current node
            node_data.branch_affiliation = copy.deepcopy(branch_stack)

            # mark children
            for child in get_children(self.graph, node):
                mark_branched_section(child, copy.deepcopy(branch_stack))

            # mark successors (at most one successor can exist,
            # since parallelization suggestions have not been imported yet)
            for successor in get_successors(self.graph, node):
                mark_branched_section(successor, copy.deepcopy(branch_stack))

        for function_node in get_all_function_nodes(self.graph):
            current_node = get_children(self.graph, function_node)[0]
            mark_branched_section(current_node, [])

    def __calculate_data_flow(self):
        self.in_data_flow = dict()
        self.out_data_flow = dict()

        def inlined_data_flow_calculation(current_node, current_last_writes):
            while current_node is not None:
                # check if current_node uses written data
                reads, writes = get_read_and_written_data_from_subgraph(
                    self.graph, current_node, ignore_successors=True
                )

                for mem_reg in current_last_writes:
                    # check if incoming data flow exists
                    if mem_reg in reads or mem_reg in writes:
                        # uses written data
                        # create incoming data flow
                        if current_node not in self.in_data_flow:
                            self.in_data_flow[current_node] = set()
                        self.in_data_flow[current_node].add(current_last_writes[mem_reg])
                        # create outgoing data flow
                        if current_last_writes[mem_reg] not in self.out_data_flow:
                            self.out_data_flow[current_last_writes[mem_reg]] = set()
                        self.out_data_flow[current_last_writes[mem_reg]].add(current_node)
                # save write to calculate outgoing data flow
                for mem_reg in writes:
                    current_last_writes[mem_reg] = current_node

                # inline children
                for child in get_children(self.graph, current_node):
                    current_last_writes = inlined_data_flow_calculation(child, current_last_writes)

                # continue to successor
                successors = get_successors(self.graph, current_node)
                if len(successors) > 1:
                    raise ValueError("only a single successor should exist at this stage in the process!")
                elif len(successors) == 1:
                    current_node = successors[0]
                else:
                    current_node = None

            return current_last_writes

        # Note: at this point in time, the graph MUST NOT have branched sections
        for function_node in get_all_function_nodes(self.graph):
            last_writes: Dict[MemoryRegion, int] = dict()
            inlined_data_flow_calculation(get_children(self.graph, function_node)[0], last_writes)

        for key in self.out_data_flow:
            for entry in self.out_data_flow[key]:
                if not self.graph.has_edge(key, entry):
                    add_temporary_edge(self.graph, key, entry)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
import os
from typing import Optional

from ..PETGraphX import LoopNode, Node, NodeID, LineID, PETGraphX
from ..utils import calculate_workload, calculate_per_iteration_workload_of_loop


class PatternInfo(object):
    """Base class for pattern detection info"""

    pattern_id: int
    _node: Node
    node_id: NodeID
    start_line: LineID
    end_line: LineID
    iterations_count: int
    average_iteration_count: int
    entries: int
    instructions_count: Optional[int]
    workload: Optional[int]
    per_iteration_workload: Optional[int]
    dp_optimizer_device_id: Optional[int] = None  # used by discopop_optimizer. unused by discopop_explorer.

    def __init__(self, node: Node):
        """
        :param node: node, where pipeline was detected
        """
        # use blocking file i/o to synchronize threads
        with open(os.path.join(os.getcwd(), "next_free_pattern_id.txt"), "r+") as f:
            lines = f.readlines()
            f.truncate(0)
            f.seek(0)
            if len(lines) == 0:
                self.pattern_id = 0
                f.write(str(0))
            else:
                for line in lines:
                    line = line.replace("\n", "").replace("\x00", "")
                    self.pattern_id = int(line)
                    f.write(str(self.pattern_id + 1))
        self._node = node
        self.node_id = node.id
        self.start_line = node.start_position()
        self.end_line = node.end_position()
        self.average_iteration_count = (
            node.loop_data.average_iteration_count
            if (isinstance(node, LoopNode) and node.loop_data is not None)
            else -1
        )
        self.iterations_count = (
            node.loop_data.total_iteration_count if (isinstance(node, LoopNode) and node.loop_data is not None) else -1
        )
        self.entries = node.loop_data.entry_count if (isinstance(node, LoopNode) and node.loop_data is not None) else -1

        # TODO self.instructions_count = total_instructions_count(pet, node)
        self.workload = None
        self.per_iteration_workload: Optional[int] = None
        # TODO self.workload = calculate_workload(pet, node)

    def to_json(self):
        dic = self.__dict__
        keys = [k for k in dic.keys()]
        for key in keys:
            if key.startswith("_"):
                del dic[key]

        return json.dumps(dic, indent=2, default=lambda o: o.toJSON())  # , default=lambda o: "<not serializable>")

    def get_workload(self, pet: PETGraphX) -> int:
        """returns the workload of self._node"""
        if self.workload is not None:
            return self.workload
        self.workload = calculate_workload(pet, self._node)
        return self.workload

    def get_per_iteration_workload(self, pet: PETGraphX) -> int:
        """returns the per iteration workload of self"""
        if self.per_iteration_workload is not None:
            return self.per_iteration_workload
        self.per_iteration_workload = calculate_per_iteration_workload_of_loop(pet, self._node)
        return self.per_iteration_workload



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import os
import random
import shutil
import statistics
import string
import subprocess
import time
import warnings
from pathlib import Path
from typing import Dict, cast, List, TextIO, Tuple, Any

import jsonpickle  # type: ignore

from discopop_library.PathManagement.PathManagement import load_file_mapping
from discopop_library.discopop_optimizer.bindings.CodeStorageObject import CodeStorageObject


def execute_stored_models(arguments: Dict[str, Any]):
    """Collects and executes all models stored in the current project path"""
    print("Cleaning environment...")
    __initialize_measurement_directory(arguments)
    print("Executing stored models...")

    # collect models to be executed
    working_copy_dir = os.path.join(arguments["--project"], ".discopop_optimizer_code_copy")
    for file_name in os.listdir(str(arguments["--code-export-path"])):
        print("\t", file_name)
        __create_project_copy(arguments["--project"], working_copy_dir)
        code_modifications = __load_code_storage_object(os.path.join(str(arguments["--code-export-path"]), file_name))
        __apply_modifications(
            arguments["--project"],
            working_copy_dir,
            code_modifications,
            load_file_mapping(arguments["--file-mapping"]),
        )
        __compile(arguments, working_copy_dir, arguments["--compile-command"])
        __measure_and_execute(arguments, working_copy_dir, code_modifications)
        # __cleanup(working_copy_dir)


def execute_single_model(arguments: Dict[str, Any]):
    """Executes the single models specified by the given arguments"""
    print("Cleaning environment...")
    __initialize_measurement_directory(arguments)

    print("Executing stored model...")

    # collect model to be executed
    working_copy_dir = os.path.join(arguments["--project"], ".discopop_optimizer_code_copy")
    file_name = arguments["--execute-single-model"]
    print("\t", file_name)
    __create_project_copy(arguments["--project"], working_copy_dir)
    code_modifications = __load_code_storage_object(os.path.join(str(arguments["--code-export-path"]), file_name))
    __apply_modifications(
        arguments["--project"],
        working_copy_dir,
        code_modifications,
        load_file_mapping(arguments["--file-mapping"]),
    )
    __compile(arguments, working_copy_dir, arguments["--compile-command"])
    __measure_and_execute(arguments, working_copy_dir, code_modifications)
    # __cleanup(working_copy_dir)


def __initialize_measurement_directory(arguments: Dict[str, Any]):
    measurement_dir = os.path.join(arguments["--project"], ".discopop_optimizer_measurements")
    if not arguments["--execution-append-measurements"]:
        # delete measurement directory
        if os.path.exists(measurement_dir):
            shutil.rmtree(measurement_dir)
    if not os.path.exists(measurement_dir):
        os.makedirs(measurement_dir)
    __initialize_measurement_file(os.path.join(measurement_dir, "measurements.csv"))


def __initialize_measurement_file(measurement_file: str):
    if not os.path.exists(measurement_file):
        with open(measurement_file, "w+") as f:
            # write file header
            header_line = "Test_case_id;Model_ID;Model_Label;return_code;Executable_name;Executable_arguments;execution_time;Function_name;\n"
            f.write(header_line)


def __measure_and_execute(arguments: Dict[str, Any], working_copy_dir: str, code_mod_object: CodeStorageObject):
    """Setup measurements, execute the compiled program and output the measurement results to a file"""
    measurement_dir = os.path.join(arguments["--project"], ".discopop_optimizer_measurements")
    # create output file for specific model measurement
    measurement_file = os.path.join(measurement_dir, "measurements.csv")

    with open(measurement_file, "a") as f:
        execution_times: List[float] = []
        test_case_id = "".join(random.choices(string.ascii_uppercase + string.digits, k=8))
        for execution_idx in range(0, int(arguments["--execution-repetitions"])):
            return_code, start_time, end_time = __execute(arguments, working_copy_dir, f)
            execution_time = end_time - start_time
            execution_times.append(execution_time)
            # write execution result
            execution_line = (
                test_case_id
                + ";"
                + code_mod_object.model_id
                + ";"
                + code_mod_object.label
                + ";"
                + str(return_code)
                + ";"
                + arguments["--executable-name"]
                + ";"
                + arguments["--executable-arguments"]
                + ";"
                + str(execution_time).replace(".", ",")
                + ";"
                + code_mod_object.parent_function.name
                + "\n"
            )
            f.write(execution_line)
            f.flush()

        print("\t\t\tREPS: ", len(execution_times))
        print("\t\t\tAVG: ", sum(execution_times) / len(execution_times))
        if len(execution_times) >= 2:
            print("\t\t\tVariance: ", statistics.variance(execution_times))


def __execute(arguments: Dict[str, Any], working_copy_dir: str, measurements_file: TextIO) -> Tuple[int, float, float]:
    """Executes the current model and returns the exit code as well as the start and end time of the execution"""
    print("\t\texecuting...")
    command = ["./" + arguments["--executable-name"], arguments["--executable-arguments"]]
    clean_command = [c for c in command if len(c) != 0]
    start_time = time.time()
    try:
        result = subprocess.run(
            clean_command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            cwd=working_copy_dir,
        )
        end_time = time.time()
        if str(result.returncode) != "0":
            warnings.warn("ERROR DURING EXECUTION...\n" + result.stderr)
        print("STDOUT: ")
        print(result.stdout)
        print("STDERR: ")
        print(result.stderr)
        return result.returncode, start_time, end_time
    except FileNotFoundError as fnfe:
        end_time = time.time()
        print(fnfe)
        return 1, start_time, end_time


def __compile(arguments: Dict[str, Any], working_copy_dir, compile_command):
    print("\t\tbuilding...")
    # command = compile_command
    # command = shlex.split(compile_command)
    if len(arguments["--make-flags"]) != 0:
        print("MAKE FLAGS: ", arguments["--make-flags"])
        compile_command += (
            " " + arguments["--make-flags"]
        )  # shlex.split(arguments["--make-flags"])  # split string, consider quotes

    if len(arguments["--make-target"]) != 0:
        compile_command += (
            " " + arguments["--make-target"]
        )  # shlex.split(arguments["--make-target"])  # split string, consider quotes
    #    clean_command = [c for c in command if len(c) != 0]
    print("\t\t\tCommand: ", compile_command)  # shlex.join(clean_command))
    result = subprocess.run(
        compile_command,
        #        " ".join(clean_command),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
        cwd=working_copy_dir,
        shell=True,
    )
    print("STDOUT: ")
    print(result.stdout)
    print("STDERR: ")
    print(result.stderr)
    if str(result.returncode) != "0":
        warnings.warn("ERROR / WARNING DURING Compilation...\n" + result.stderr)


def __apply_modifications(
    project_folder,
    working_copy_dir,
    modifications: CodeStorageObject,
    file_mapping: Dict[int, Path],
):
    print("\t\tApplying code modifications...")
    print("\t\t\tFunction: ", modifications.parent_function.name)
    for file_id in modifications.patches:
        file_mapping_path = str(file_mapping[int(file_id)])
        # remove /.discopop/ from path if it occurs
        if "/.discopop/" in file_mapping_path:
            file_mapping_path = file_mapping_path.replace("/.discopop/", "/")
        # get file to be overwritten
        replace_path = file_mapping_path.replace(project_folder, working_copy_dir)
        # apply patch to the file
        if not os.path.exists(replace_path):
            raise FileNotFoundError(replace_path)
        # save patch to disk
        patch = replace_path + ".patch"
        with open(patch, "w+") as p:
            p.write(modifications.patches[file_id])
            p.flush()
            p.close()

        # check existence of files:
        if os.path.exists(patch):
            print("PATCH FILE EXISTS")
        if os.path.exists(replace_path):
            print("REPLACE PATH EXISTS")

        # command = ["patch", "-i", patch, "-o", replace_path]
        command = ["patch", replace_path, patch]
        print("Patch command: ", command)

        result = subprocess.run(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            cwd=os.getcwd(),
        )
        print("RESULT: ", result.returncode)
        print("STDERR:")
        print(result.stderr)
        print("STDOUT: ")
        print(result.stdout)

        if result.returncode == 0:
            print("Applied Patch:")
            print(modifications.patches[file_id])


# remove temporary patch file
# todo re-enable cleanup


#        if os.path.exists(patch):
#            os.remove(patch)


def __load_code_storage_object(file_path) -> CodeStorageObject:
    json_contents = ""
    with open(file_path, "r") as f:
        json_contents = f.read()
    return cast(CodeStorageObject, jsonpickle.decode(json_contents))


def __create_project_copy(source, target):
    print("\t\tCreating a working copy of the project...", end="")
    if os.path.exists(target):
        shutil.rmtree(target)
    shutil.copytree(str(source), target, ignore=shutil.ignore_patterns(".discopop*"))
    print("Done")


def __cleanup(working_copy_dir: str):
    # remove working copy to free space
    print("\t\tRemoving working copy of the project...", end="")
    shutil.rmtree(working_copy_dir)
    print("Done.")



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import NewType

MemoryRegion = NewType("MemoryRegion", str)

VarName = NewType("VarName", str)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import os.path
from argparse import ArgumentParser
from pathlib import Path

from discopop_library.PatchGenerator.PatchGeneratorArguments import PatchGeneratorArguments
from discopop_library.PatchGenerator.patch_generator import run


def parse_args() -> PatchGeneratorArguments:
    """Parse the arguments passed to the discopop_explorer"""
    parser = ArgumentParser(description="DiscoPoP Patch Generator")
    # all flags that are not considered stable should be added to the experimental_parser
    experimental_parser = parser.add_argument_group(
        "EXPERIMENTAL",
        "Arguments for the task pattern detector and other experimental features. These flags are considered EXPERIMENTAL and they may or may not be removed or changed in the near future.",
    )

    # fmt: off
    parser.add_argument("--verbose", action="store_true",
        help="Enable verbose output.")
    parser.add_argument(
        "--dp-build-path", type=str, required=True,
        help="Path to DiscoPoP build folder"
    )
    # EXPERIMENTAL FLAGS:
    # fmt: on

    arguments = parser.parse_args()

    # determine CC and CXX
    with open(os.path.join(arguments.dp_build_path, "build_config.txt"), "r") as f:
        for line in f.readlines():
            if line.startswith("LLVM_BIN_DIR="):
                line = line.replace("\n", "")
                llvm_bin_dir = line[len("LLVM_BIN_DIR=") :]
                # determine CC
                if os.path.exists(os.path.join(llvm_bin_dir, "clang")):
                    arguments.cc = os.path.join(llvm_bin_dir, "clang")
                elif os.path.exists(os.path.join(llvm_bin_dir, "clang-11")):
                    arguments.cc = os.path.join(llvm_bin_dir, "clang-11")
                else:
                    raise ValueError("Could not determine CC from discopop build path: ", arguments.dp_build_path)

                # determine CXX
                if os.path.exists(os.path.join(llvm_bin_dir, "clang++")):
                    arguments.cxx = os.path.join(llvm_bin_dir, "clang++")
                elif os.path.exists(os.path.join(llvm_bin_dir, "clang++-11")):
                    arguments.cxx = os.path.join(llvm_bin_dir, "clang++-11")
                else:
                    raise ValueError("Could not determine CXX from discopop build path: ", arguments.dp_build_path)

                break

    return PatchGeneratorArguments(
        verbose=arguments.verbose, discopop_build_path=arguments.dp_build_path, CC=arguments.cc, CXX=arguments.cxx
    )


def main():
    arguments = parse_args()
    run(arguments)


if __name__ == "__main__":
    main()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from discopop_library.discopop_optimizer.classes.types.Aliases import DeviceID
from discopop_library.discopop_optimizer.classes.types.DataAccessType import WriteDataAccess


class Update(object):
    source_node_id: int
    target_node_id: int
    source_device_id: DeviceID
    target_device_id: DeviceID
    write_data_access: WriteDataAccess
    is_first_data_occurrence: bool

    def __init__(
        self,
        source_node_id: int,
        target_node_id: int,
        source_device_id: DeviceID,
        target_device_id: DeviceID,
        write_data_access: WriteDataAccess,
        is_first_data_occurrence: bool,
    ):
        self.source_node_id = source_node_id
        self.target_node_id = target_node_id
        self.source_device_id = source_device_id
        self.target_device_id = target_device_id
        self.write_data_access = write_data_access
        self.is_first_data_occurrence = is_first_data_occurrence

    def __str__(self):
        result_str = "First" if self.is_first_data_occurrence else ""
        return (
            result_str
            + "Update("
            + str(self.source_node_id)
            + "@"
            + str(self.source_device_id)
            + ", "
            + str(self.target_node_id)
            + "@"
            + str(self.target_device_id)
            + ", "
            + " : "
            + str(self.write_data_access.memory_region)
            + " --> "
            + str(self.write_data_access.var_name)
            + ")"
        )



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

"""Discopop Code Generator

Usage:
    discopop_code_generator --fmap <path> --json <path> --outputdir <path> [--patterns <str>] [--compile-check-command <str>] [--skip-compilation-check]

OPTIONAL ARGUMENTS:
    --fmap=<file>               File mapping
    --json=<file>               Json output of the DiscoPoP Explorer
    --outputdir=<path>          Directory for the modified source files
    --patterns=<str>            Comma-separated list of pattern types to be applied
                                Possible values: reduction, do_all, simple_gpu, combined_gpu
    --skip-compilation-check    Do not validate the inserted patterns by compiling the resulting source code.
    --compile-check-command=<str>     Specify a command to be executed for performing the compile check (e.g. "cd .. && make")
    -h --help                   Show this screen
"""
import os
import sys
from typing import Dict, List

from docopt import docopt  # type:ignore
from schema import Schema, Use, SchemaError  # type:ignore

from discopop_library.CodeGenerator.CodeGenerator import (
    from_json_strings as generate_code_from_json_strings,
)
from discopop_library.JSONHandler.JSONHandler import read_patterns_from_json_to_json
from discopop_library.PathManagement.PathManagement import load_file_mapping, get_path

docopt_schema = Schema(
    {
        "--fmap": Use(str),
        "--json": Use(str),
        "--patterns": Use(str),
        "--outputdir": Use(str),
        "--skip-compilation-check": Use(str),
        "--compile-check-command": Use(str),
    }
)


def main():
    """Applies the code modifications specified by the given JSON file and writes
    the modified source code structure into outputdir.
    Note that only modified source code files will be written to outputdir."""
    arguments = docopt(__doc__)

    try:
        arguments = docopt_schema.validate(arguments)
    except SchemaError as e:
        exit(e)

    file_mapping_file = get_path(os.getcwd(), arguments["--fmap"])
    json_file = get_path(os.getcwd(), arguments["--json"])
    outputdir = arguments["--outputdir"]
    relevant_patterns: List[str] = (
        []
        if arguments["--patterns"] == "None"
        else (arguments["--patterns"].split(",") if "," in arguments["--patterns"] else [arguments["--patterns"]])
    )
    # validate patterns
    for pattern in relevant_patterns:
        if pattern not in ["reduction", "do_all", "simple_gpu", "combined_gpu"]:
            raise ValueError("Unsupported pattern name: ", pattern, " given in '--patterns' argument!")

    for file in [file_mapping_file, json_file]:
        if not os.path.isfile(file):
            raise FileNotFoundError("File not found: ", file)
    if not os.path.exists(outputdir):
        os.mkdir(outputdir)

    file_mapping_dict = load_file_mapping(file_mapping_file)

    identified_patterns = read_patterns_from_json_to_json(json_file, relevant_patterns)

    print("SKIP COMPILATION CHECK? ", arguments["--skip-compilation-check"], file=sys.stderr)
    print("\ttype: ", type(arguments["--skip-compilation-check"]), file=sys.stderr)
    print("COMPILE CHECK COMMAND? ", arguments["--compile-check-command"], file=sys.stderr)
    print("\ttype: ", type(arguments["--compile-check-command"]), file=sys.stderr)

    print("FILE MAPPING: ", file_mapping_dict)
    print("patterns:", identified_patterns)

    modified_code = generate_code_from_json_strings(
        file_mapping_dict,
        identified_patterns,
        skip_compilation_check=True if arguments["--skip-compilation-check"] != "False" else False,
        compile_check_command=arguments["--compile-check-command"]
        if arguments["--compile-check-command"] != "None"
        else None,
    )

    modified_code_by_new_location: Dict[str, str] = dict()
    for file_id in modified_code:
        old_location = file_mapping_dict[file_id]
        print("OLD LOC: ", old_location)
        new_location = os.path.join(outputdir, str(old_location).split("/.discopop/")[1])
        print("NEW LOC: ", new_location)

        modified_code_by_new_location[new_location] = modified_code[file_id]

    # output modified code
    for file_path in modified_code_by_new_location:
        with open(file_path, "w+") as f:
            f.write(modified_code_by_new_location[file_path])
            f.close()

    print("Written modified source code to: ", outputdir)


if __name__ == "__main__":
    main()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import warnings
from typing import Dict, List, Tuple, Optional

from sympy import Symbol, Expr, Integer

from discopop_library.discopop_optimizer.classes.system.Network import Network
from discopop_library.discopop_optimizer.classes.system.devices.CPU import CPU
from discopop_library.discopop_optimizer.classes.system.devices.Device import Device
from discopop_library.discopop_optimizer.classes.system.devices.GPU import GPU


class System(object):
    __devices: Dict[int, Device]
    __network: Network
    __next_free_device_id: int
    __device_do_all_overhead_models: Dict[Device, Expr]
    __device_reduction_overhead_models: Dict[Device, Expr]

    def __init__(self, headless: bool = False):
        self.__devices = dict()
        self.__network = Network()
        self.__next_free_device_id = 0
        self.__device_do_all_overhead_models = dict()
        self.__device_reduction_overhead_models = dict()

        # define a default system
        # todo replace with benchmark results and / or make user definable
        if headless:
            device_0_threads = Integer(4)
        else:
            device_0_threads = Symbol("device_0_threads")  # Integer(48)
        device_0 = CPU(
            Integer(3000000000),
            device_0_threads,
            openmp_device_id=-1,
            device_specific_compiler_flags="COMPILE FOR CPU",
        )  # Device 0 always acts as the host system
        gpu_compiler_flags = "COMPILE FOR CPU"
        device_1 = GPU(
            Integer(512000000),
            Integer(512),
            openmp_device_id=0,
            device_specific_compiler_flags="COMPILE FOR GPU",
        )
        device_2 = GPU(
            Integer(512000000),
            Integer(512),
            openmp_device_id=1,
            device_specific_compiler_flags="COMPILE FOR GPU",
        )
        self.add_device(device_0)
        self.add_device(device_1)
        self.add_device(device_2)
        # define Network
        network = self.get_network()
        network.add_connection(device_0, device_0, Integer(100000), Integer(0))
        network.add_connection(device_0, device_1, Integer(100), Integer(1000000))
        network.add_connection(device_1, device_0, Integer(100), Integer(1000000))
        network.add_connection(device_1, device_1, Integer(100000), Integer(0))

        network.add_connection(device_0, device_2, Integer(100), Integer(10000000))
        network.add_connection(device_2, device_0, Integer(100), Integer(10000000))
        network.add_connection(device_2, device_2, Integer(1000), Integer(0))

        network.add_connection(device_1, device_2, Integer(100), Integer(500000))
        network.add_connection(device_2, device_1, Integer(100), Integer(500000))

    # todo: support the replication of device ids (e.g. CPU-0 and GPU-0)

    def set_device_doall_overhead_model(self, device: Device, model: Expr):
        print("System: Set DOALL overhead model: ", model)
        self.__device_do_all_overhead_models[device] = model

    def set_reduction_overhead_model(self, device: Device, model: Expr):
        print("System: Set REDUCTION overhead model: ", model)
        self.__device_reduction_overhead_models[device] = model

    def get_device_doall_overhead_model(self, device: Device) -> Expr:
        if device not in self.__device_do_all_overhead_models:
            warnings.warn("No DOALL overhead model, assuming 0 for device: " + str(device))
            return Expr(Integer(0))
        return self.__device_do_all_overhead_models[device]

    def get_device_reduction_overhead_model(self, device: Device) -> Expr:
        if device not in self.__device_reduction_overhead_models:
            warnings.warn("No REDUCTION overhead model, assuming 0 for device: " + str(device))
            return Expr(Integer(0))
        return self.__device_reduction_overhead_models[device]

    def add_device(self, device: Device):
        device_id = self.__next_free_device_id
        self.__next_free_device_id += 1
        self.__devices[device_id] = device

    def get_device(self, device_id: Optional[int]) -> Device:
        if device_id is None:
            return self.__devices[0]
        return self.__devices[device_id]

    def get_network(self) -> Network:
        return self.__network

    def get_free_symbols(self) -> List[Tuple[Symbol, Optional[Expr]]]:
        result_list: List[Tuple[Symbol, Optional[Expr]]] = []
        for device in self.__devices.values():
            result_list += device.get_free_symbols()
        result_list += self.__network.get_free_symbols()
        return result_list

    def get_device_id(self, device: Device) -> int:
        for key in self.__devices:
            if device == self.__devices[key]:
                return key
        raise ValueError("Unknown device: ", device)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import List


class Arguments(object):
    no_gui: bool = False
    execute_all_configurations: bool = False
    execute_configurations_with_tag: List[str] = []



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from sympy import Symbol  # type: ignore

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.classes.edges.GenericEdge import GenericEdge


class OptionEdge(GenericEdge):
    """Used to mark selection options for the original sequential version.
    Used in combination with requirement edges to restrain path selections."""

    pass

    def get_cost_model(self) -> CostModel:
        raise ValueError("The cost of an OptionEdge is not defined and may never be used!")



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Dict, List, Tuple, cast

import networkx as nx  # type: ignore

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.nodes.ContextNode import ContextNode
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.utilities.MOGUtilities import (
    get_successors,
    data_at,
    get_children,
)


def calculate_data_transfers(
    graph: nx.DiGraph, function_performance_models: Dict[FunctionRoot, List[CostModel]]
) -> Dict[FunctionRoot, List[Tuple[CostModel, ContextObject]]]:
    """Calculate data transfers for each performance model and append the respective ContextObject to the result."""
    result_dict: Dict[FunctionRoot, List[Tuple[CostModel, ContextObject]]] = dict()
    for function in function_performance_models:
        result_dict[function] = []
        for model in function_performance_models[function]:
            # create a ContextObject for the current path
            context = ContextObject(function.node_id, [function.device_id])
            context = get_path_context(function.node_id, graph, model, context)
            result_dict[function].append((model, context))
    return result_dict


def get_path_context(node_id: int, graph: nx.DiGraph, model: CostModel, context: ContextObject) -> ContextObject:
    """passes the context Object along the path and returns the context once the end has been reached"""
    # push device id to stack if necessary
    node_data = data_at(graph, node_id)
    if node_data.device_id is not None:
        context.last_seen_device_ids.append(node_data.device_id)

    # calculate context modifications for the current node
    context = __check_current_node(node_id, graph, model, context)

    # calculate context modifications for the children of the current node
    context = __check_children(node_id, graph, model, context)

    # pop device id from stack if necessary
    if node_data.device_id is not None:
        context.last_seen_device_ids.pop()

    # set last_visited_node_id to the original node_id,
    # since the calculation continues from node_id after the children have been visited
    context.last_visited_node_id = node_id

    # pass context to the right successor of the current node
    # At most a single node can be a successor, since the given model represents a single path through the graph.
    successors = get_successors(graph, node_id)
    if len(successors) == 1:
        # pass context to the single successor
        return get_path_context(successors[0], graph, model, context)

    elif len(successors) == 0:
        # no successor exists, return the current context
        return context

    else:
        # multiple successors exist
        # find the successor which represents the path decision included in the model
        suitable_successors = [succ for succ in successors if succ in model.path_decisions]
        if len(suitable_successors) != 1:
            raise ValueError(
                "Invalid amount of potential successors (",
                len(suitable_successors),
                ") for path split at node:",
                node_id,
                "using decisions: ",
                model.path_decisions,
                "successors:",
                successors,
            )
        # suitable successor identified.
        # pass the current context to the successor
        return get_path_context(suitable_successors[0], graph, model, context)


def __check_current_node(node_id: int, graph: nx.DiGraph, model: CostModel, context: ContextObject) -> ContextObject:
    """Check if the given node results in modifications to the given context.
    Return a modified version of the context which contains the required updates."""
    # due to the Read-Compute-Write paradigm used to create the Computational Units,
    # this structure is assumed for the nodes and their MemoryAccesses as well.

    # check if any data needs to be updated from a different device before reading
    # if so, the context will be supplemented with the identified updates
    node_data = data_at(graph, node_id)

    if isinstance(node_data, ContextNode):
        # apply modifications according to encountered context node
        updated_context = cast(ContextNode, data_at(graph, node_id)).get_modified_context(
            node_id, graph, model, context
        )
        return updated_context

    context = context.calculate_and_perform_necessary_updates(
        node_data.read_memory_regions,
        cast(int, context.last_seen_device_ids[-1]),
        node_data.node_id,
    )

    # add the writes performed by the given node to the context
    context = context.add_writes(node_data.written_memory_regions, cast(int, context.last_seen_device_ids[-1]))

    return context


def __check_children(node_id: int, graph: nx.DiGraph, model: CostModel, context: ContextObject) -> ContextObject:
    # pass context to all children
    for child in get_children(graph, node_id):
        # reset last_visited_node_id inbetween visiting children
        context.last_visited_node_id = node_id
        context = get_path_context(child, graph, model, context)
    return context
